/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * 
 */
package bit.lsi.sar.nsg;


import java.io.StringReader;
import java.util.HashMap;
import java.util.HashSet;

import joelib.io.IOTypeHolder;
import joelib.molecule.JOEMol;
import joelib.smiles.JOESmilesParser;

import org.openscience.cdk.ChemFile;
import org.openscience.cdk.interfaces.IChemModel;
import org.openscience.cdk.interfaces.IChemSequence;
import org.openscience.cdk.interfaces.IMolecule;
import org.openscience.cdk.interfaces.IMoleculeSet;
import org.openscience.cdk.io.MDLReader;
import org.openscience.cdk.io.Mol2Reader;
import org.openscience.cdk.layout.StructureDiagramGenerator;

import cern.colt.bitvector.BitVector;


/**
 * @author lounkine
 *
 */
public class Molecule implements Comparable<Molecule> {

	/**all
	 * 
	 */
	//private static final long serialVersionUID = -8785345949591819768L;
	private static int molCount = 0;
	private static StructureDiagramGenerator sdg = new StructureDiagramGenerator();
	
	private int internalId = 0;
	
	private String externalId;
	private String smiles;
	private String name;
	private BitVector fingerprint;
	//private JOEMol joeMol;
	private HashMap<String, String> descriptors = new HashMap<String,String>();
	
	private HashMap<Target, Potency> potencies = new HashMap<Target, Potency>();
	
	private IMolecule cdkmol = null;
	private boolean converted = false;
	
	
	private static String removeChar(String s, char c) {
		   String r = "";
		   for (int i = 0; i < s.length(); i ++) {
		      if (s.charAt(i) != c) r += s.charAt(i);
		      }
		   return r;
	}
	
	
	/**
	 * 
	 * @param smiles
	 * @param externalId
	 * @param name
	 * @param fpstring
	 */
	public Molecule(String smiles, String externalId, String name, BitVector fp) {
		
		
		this.smiles = smiles;
		this.externalId = externalId;
		this.name = name;
		this.internalId = ++molCount;
		this.fingerprint = fp;		
	}
	
	private boolean convertMolecule() {
		
		JOEMol joeMol = new JOEMol(IOTypeHolder.instance().getIOType("SMILES"),
				IOTypeHolder.instance().getIOType("MOL2"));
		
	
		
		if (JOESmilesParser.smiToMol(joeMol, smiles, name))
		{
			  
		  Mol2Reader reader = new Mol2Reader(new StringReader(joeMol.toString()));
        
			ChemFile chemfile;
				try {
					chemfile = (ChemFile) reader.read(new ChemFile());
				
		        IChemSequence seq = chemfile.getChemSequence(0);
		        IChemModel model = seq.getChemModel(0);
		        IMoleculeSet moleculeSet = model.getMoleculeSet();
		        cdkmol = moleculeSet.getMolecule(0);
		        //CDKHueckelAromaticityDetector.detectAromaticity(mol);
		        
		        
		      
		         sdg.setMolecule((IMolecule)cdkmol.clone());
		         sdg.generateCoordinates();
		         cdkmol = sdg.getMolecule();
		                  
		           
		            
				}
		        catch(Exception exc)
		        {
		            System.err.println("*** Exit due to an unexpected error during coordinate generation ***");
		            exc.printStackTrace();
		        }
		        
		                
			
			
		}
		else
		{
			System.err.println("Molecule could not be generated from \"" + smiles + "\".");
			return false;
		}
		
		
		
		return true;
	}
	
	
	public IMolecule getCDKMol() {
		if (!converted)
			converted =  convertMolecule();
		return cdkmol;
	}
	
	public boolean clearCDKMol() {
		if (!converted)
			return false;
		else {
			cdkmol = null;
			converted = false;
			return true;
		}
		
	}
	
	/*
	public Molecule(JOEMol mol) {
		
		joeMol = mol;
		GenericDataIterator gdit = mol.genericDataIterator();
		JOEGenericData genericData;
		
		this.smiles = mol.getData("SMILES").toString();
		
		
		while (gdit.hasNext()) {
			
			genericData = gdit.nextGenericData();
			
			 if ( genericData.getDataType() == JOEDataType.JOE_PAIR_DATA )
			 {
				 
				 
			 }
			
			
		}
		
	}
	*/
	
	
	/** Returns the name of the molecule */
	public String toString() {
		
		return new String (name);
	}
	
	/** Returns (a copy of) the SMILES String */
	public String getSmiles() {
		
		return new String (smiles);
	}
	
	public void finalize() {
		molCount--;
	}
	
	
	/** Calculates the Tc based on molecule fingerprints.
	 * 
	 * @param mol1 Molecule 1 
	 * @param mol2 Molecule 2
	 * @return Tc Tanimoto coefficient
	 */
	public static double Tc (Molecule mol1, Molecule mol2) {
		
		double Tc = 0;
		int a, b, c;
		BitVector fp1, fp2, hfp;
		
		
		
		
		fp1 = mol1.getFingerprint();
		fp2 = mol2.getFingerprint();
		
		hfp = fp1.copy();
		
		a = fp1.cardinality();
		b = fp2.cardinality();
		
		if (a + b == 0)
			return 0;
			
		hfp.and(fp2);
		c = hfp.cardinality();
		
		Tc = ((double)c / ((double)a + (double)b - (double)c));
		
		return Tc;
	}
	
	
	/**
	 * 
	 * @param mol Molecule with which this molecule is compared.
	 * @return Tanimoto Coefficient.
	 */
	public double Tc (Molecule mol) {
		return Tc(this, mol);
	}
	
	
	/**
	 * Returns true iff the smiles strings of the two molecules are identical.
	 */
	public boolean equals(Object mol2) {
		return this.smiles.equals(((Molecule)mol2).getSmiles());
	}
	
	public int hashCode() {
		return smiles.hashCode();
	}
	
	/**
	 * 
	 * @return Fingerprint.
	 */
	public BitVector getFingerprint() {
		
		return this.fingerprint;
		
	}
	
	/**
	 * 
	 * @return Copy of the external id.
	 */
	public String getExternalId() {
		
		return new String(externalId);
	}
	
	/**
	 * 
	 * @return Internal Id (serial integer)
	 */
	public int getInternalId () {
		return internalId;
	}
	
	/**
	 * Adds a new descriptor with the specified value. If a descriptor with this name is already present,
	 * false is returned and the descriptor is not set. In order to set a new value, use setDescriptor.
	 * @param prop Name of the Descriptor (Property)
	 * @param value Descriptor Value 
	 * @return false if Descriptor with this name is already present. true if the descriptor was added successfully.
	 * 
	 */
	public boolean addDescriptor (String prop, String value) {
		
		if (descriptors.containsKey(prop))
			return false;
		else {
			descriptors.put(prop, value);
			return true;
		}
		
	}
	
	/**
	 * Adds a new Descriptor and initializes it with an empty string.
	 * @param prop Descriptor name
	 * @return true iff Descriptor was not yet present
	 */
	public boolean addDescriptor (String prop) {
		
		if (descriptors.containsKey(prop))
			return false;
		else {
			descriptors.put(prop, new String());
			return true;
		}
		
	}
	
	/**
	 * Updates descriptor value. If the descriptor is not known, returns false.
	 * @param prop Descriptor name
	 * @param value Descriptor value
	 * @return true iff Descriptor is known.
	 */
	public boolean setDescriptor (String prop, String value) {
		
		if (descriptors.containsKey(prop)){
			descriptors.put(prop, value);
			return true;
		}
		else 
			return false;
		
	}
	
	/**
	 * Gets the associated value of a descriptor.
	 * @param prop Descriptor name
	 * @return Descriptor value if the Descriptor is known, null else.
	 */
	public String getDescriptorValue (String prop) {
		
		return descriptors.get(prop);
		
	}
	
	/**
	 * Removes a descriptor.
	 * @param prop
	 * @return true iff the removal was successful, false if the descriptor is not known.
	 */
	public boolean removeDescriptor(String prop){
		if (descriptors.containsKey(prop))
		{
			descriptors.remove(prop);
			return true;
		}
		else
			return false;
	}
	
	
	public HashSet<String> getDescriptorNames () {
		
		return new HashSet<String>(descriptors.keySet());
		
	}
	
	
	public void addPotency(Target t, Potency p) {
		
		potencies.put(t,p);
		
	}
	
	public boolean removePotency (Target t) {
		
		if (potencies.containsKey(t)) {
			
			potencies.remove(t);
			t.removeMolecule(this);
			return true;
		}
		
		return false;
	}
	
	
	public Potency getPotency( Target t) {
		
		return potencies.get(t);
		
	}
	
	
	public double getpSR (Target t1, Target t2) {
		
		return this.getPotency(t1).pSR(this.getPotency(t2));
	}
	
	public HashSet<Target> getTargets() {
		
		return new HashSet<Target>(potencies.keySet());
		
	}
	
	
	public HashSet<SARSet> getAllSARSets() {
		HashSet<SARSet> set = new HashSet<SARSet>();
		
		for (Target t : this.getTargets()) {
			
			SARSet cs;
			if ((cs = t.getSARSet()) != null)		
				set.add(cs);
		}
		
		return set;
		
	}
	
	public HashSet<SSRSet> getAllSSRSets(double cutOff, HashMap<String,Double> norms) {
	
		HashSet<SSRSet> set = new HashSet<SSRSet>();
		
		for (Target t1 : this.getTargets()) {
			
			for (Target t2 : this.getTargets())
			{
			
				if (t1.compareTo(t2) >= 0)
					continue;
				
				SSRSet cs;
				if ((cs = SSRSet.getSSRSet(t1, t2)) != null)		
					set.add(cs);
			}
		}
		
		return set;
		
	}
	
	public boolean isInAddSet () {
		
		for (Target t : potencies.keySet()) {
			
			if (!t.hasPotencyAnnotation())
				return true;
			
		}
		
		return false;
		
	}


	@Override
	public int compareTo(Molecule arg0) {
		return this.smiles.compareTo(arg0.getSmiles());
		
	}
	
	
	
		
}
