/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * 
 */
package bit.lsi.sar.nsg;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

import cern.colt.matrix.DoubleFactory2D;
import cern.colt.matrix.DoubleMatrix1D;
import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix1D;



/**
 * @author lounkine
 *
 */
public class MoleculeSet extends HashSet<Molecule>{

	/**
	 * 
	 */
	private static final long serialVersionUID = -1018668413242817399L;
	
	private HashMap<Molecule, Integer> molIds = new HashMap<Molecule, Integer>();
	private DoubleMatrix2D tcMatrix = null;
	private int totalComparisons = 0;
	
	
	private void initMolIds() {
		int i = 0;
		
		for (Molecule m : this) {
			molIds.put(m, i++);
		}
	}
	
	
	public MoleculeSet() {
		super();	
	}
	
	public MoleculeSet(Collection<Molecule> c) {
		super(c);
		initMolIds();
		
		
	}
	
	public ArrayList<Double> getAllTcValues(double cutoff) {
		if (tcMatrix == null)
			precalcTcMatrix();
		
		ArrayList<Double> dd1d = new ArrayList<Double>(tcMatrix.cardinality());
		
		double curTc;
		for (Molecule m1 : this) {
			for (Molecule m2 : this) {
				if (molIds.get(m1) >= molIds.get(m2))
					continue;
				
				curTc = tcMatrix.getQuick(molIds.get(m1), molIds.get(m2));
				
				if (curTc > cutoff)
					dd1d.add(curTc);
				
							
			}
		}
		
		return dd1d;
	}
	
	public ArrayList<Double> getAllTcValues() {
		return getAllTcValues(0d);
	}
	
	
	
	
	
	/**
	 * Returns a Molecule set containing the neighbors of the current molecule and the molecule.
	 * @param mol Central Molecule
	 * @param Threshold Tc cutoff of what is considered neighbor (> cutoff)
	 * @return MoleculeSet neighbors
	 */
	public MoleculeSet getNeighbors(Molecule mol, double Threshold) {
		
		MoleculeSet curset = new MoleculeSet();
		
		for (Molecule othermol: this) {
			if (mol.equals(othermol))
				continue;
			
			if (this.getPrecalcTc(mol, othermol) > Threshold)
				curset.add(othermol);
		}
		
		return curset;
	}
	
	
	
	
	protected void precalcTcMatrix() {
		
		if (tcMatrix == null) {
		
			tcMatrix = DoubleFactory2D.sparse.make(this.size(), this.size());
			
			totalComparisons = this.size() * (this.size()-1) / 2;
			long curComp = 0;
			double curPercent;
			int i1, i2;
			
			for (Molecule m1 : this) {
				
				i1 = molIds.get(m1);
				//System.err.println(m1);
				tcMatrix.setQuick(i1, i1, 1);
				for (Molecule m2 : this) {
					i2 = molIds.get(m2);
					if (i1 >= i2)
						continue;
					
					
					
					curPercent = 100 *  ++curComp / totalComparisons;
					
					if ( curComp % 100000 == 0)
						System.err.println(curPercent + "% completed");
					
					
					tcMatrix.setQuick(i1, i2, m1.Tc(m2));				
				}
				
			}
		}
		//System.err.print("\n");
		
	
	}
	
	public double getPrecalcTc(Molecule m1, Molecule m2) {
		
			
		if (tcMatrix == null) {
			
			precalcTcMatrix();
			return  getPrecalcTc(m1, m2);
			
		}
			
		
		if (!(this.contains(m1) && this.contains(m2)))
			return m1.Tc(m2);
		
		
		int i1 = molIds.get(m1);
		int i2 = molIds.get(m2);
			
		if (i1 == i2)
			return 1;
		
		if (i1 > i2) {
			int c = i1;
			i1 = i2;
			i2 = c;
			
		}
				
		return tcMatrix.get(i1, i2);
		
	}
	
	
	public void removeAllCDKMols() {
		
		for (Molecule m: this)
			m.clearCDKMol();
		
		Runtime.getRuntime().gc();
	}
	
	public void makeAllCDKMols() {

		//for (Molecule m: this)
		//	m.getCDKMol();
	}
	
			
		
		
	
	
	
}
