/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * 
 */
package bit.lsi.sar.nsg;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;


import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.SparseGraph;



import edu.uci.ics.jung.graph.util.EdgeType;






/**
 * @author lounkine
 *
 */
public class NSG{
	
	/**
	 * 
	 */
	//private static final long serialVersionUID = 2671439762112286312L;

	public class NSGVertex {
		
		/**
		 * 
		 */
		//private static final long serialVersionUID = -6209186830777565577L;
		private Molecule mol;
		private double sizeV;
		private double colorV;
		private boolean inSarSet = false;
		private Color lastSetColor = Color.black;
		private double lastThreshold;
		
		
		private boolean hasAddNeighbor = false;
		
		private Set<MoleculeSet> setsOfVertex = new HashSet<MoleculeSet>();;
		
		
		public void setSetColor(Color col) {
			this.lastSetColor = col;
		}
		
		public Color getSetColor() {
			return this.lastSetColor;
		}
		
		
		private double normColorValue(double cV) {
			
			//System.err.println("min: " + minProp + "\nmax: " + maxProp + "\ncv: " + cV);
			
			return (cV - minProp) / (maxProp - minProp);
			
		}
		
		public NSGVertex (Molecule m, double sizeValue, double colorValue, boolean inSARSet) {
			
			
			//System.err.println(colorValue + "  " + normColorValue(colorValue));
			
			this.mol = m;
			this.sizeV = Math.sqrt(sizeValue);
			this.colorV = normColorValue(colorValue);
			this.inSarSet = inSARSet;
		}
		
		public void addToSet(MoleculeSet ms) {
			setsOfVertex.add(ms);
		}
		
		public void removeFromSet(MoleculeSet ms) {
			setsOfVertex.remove(ms);
		}
		
		public boolean isInSet(MoleculeSet ms) {
			return setsOfVertex.contains(ms);
		}
		
		public boolean isInAddSet() {
			return (this.setsOfVertex.size() > 0);
		}
		
		public Molecule getMol() {
			return this.mol;
		}
		
		public boolean isInSarSet() {
			return this.inSarSet;
		}
		
		
		public double getSizeValue() {
			
			if (inSarSet) {
				
				return this.sizeV;
				
			}
			else
			{
				return (double) ((double) nsGraph.degree(this) / sarset.getNumberMolecules());
			}
		}
		
		public double getCentralProperty() {
			return sarset.getCentralProperty(this.getMol());
		}
		
		public String getCentralPropertyText() {
			return sarset.getCentralPropertyForTable(this.getMol());
		}
		
		
		
		public double getColorValue() {
			
				
				return this.colorV;
				
			
		}
		
		public String toString () {
			return this.mol.toString();
		}
		
		public void setThreshold(double t) {
			lastThreshold = t;
		}
		
		public double getThreshold () {
			return lastThreshold;
		}

		public void setHasAddNeighbor(boolean hasAddNeighbor) {
			this.hasAddNeighbor = hasAddNeighbor;
		}

		public boolean hasAddNeighbor() {
			return hasAddNeighbor;
		}
		
		public boolean isSSR() {
			return (sarset.getCentralPropertyName().equals("Selectivity"));
		}
		
		
		
	}

	public class NSGEdge{
		
		/**
		 * 
		 */
		//private static final long serialVersionUID = 6359879025030320331L;
		
		private double fancyMetric;
		
		private boolean connectsAdditional = false;
		private boolean hasZeroDiff = false;
		
		private NSGVertex v1;
		private NSGVertex v2;
		
		private static final int aValue = 2;
		
		private double sim;
		
		
		public double getCentralPropDiff() {
			return Math.abs
			(getCentralProperty(v1)
			- getCentralProperty(v2) );
		}
		
		private boolean calcFancy() {
			
			if (connectsAdditional) {
				fancyMetric = Double.POSITIVE_INFINITY;
				return false;
			}
			
			fancyMetric = this.getCentralPropDiff();
			
		if (fancyMetric == 0 ) {
		
			fancyMetric = Double.POSITIVE_INFINITY;
			hasZeroDiff = true;
			return false;
		}
		
		fancyMetric = Math.pow(fancyMetric, aValue);
		
		fancyMetric /= sarset.getPrecalcTc(v1.getMol(), v2.getMol());
			return true;
		}
		
		
		public boolean hasZeroDiff() {
			return hasZeroDiff;
		}
		
		public NSGEdge(NSGVertex v1, NSGVertex v2) {
					
			this.v1 = v1;
			this.v2 = v2;
			
			
			sim = sarset.getPrecalcTc(v1.getMol(), v2.getMol());
			connectsAdditional = (v1.isInAddSet() || v2.isInAddSet());
			
			calcFancy();
			
			
							
		}
		
		
		
		
		
		public double getSim() {
			return sim;
		}
		
		public double getFancy() {
			return this.fancyMetric;
		}
		
		public String toString() {
			return "Tc =" + String.format("%.2f", this.getSim());
		}
		
		public boolean connectsAdd() {
			return connectsAdditional;
		}
		
		public NSGVertex[] getVertices() {
			NSGVertex[] vs = new NSGVertex[2];
			vs[0] = this.v1;
			vs[1] = this.v2;
			return vs;
		}
		
		
		/**
		 * If one of the vertices is in an additional set, returns it.
		 * The order is arbitrary.
		 * 
		 * @return Vertex that is in an additional set. If none is in additional set, return first vertex
		 */
		public NSGVertex getAddVertex() {
			
			if (v1.isInAddSet())
				return v1;
			else
				return v2;
			
			
		}
		
		
	}
	
	
	
	private Graph<NSGVertex, NSGEdge> nsGraph = new SparseGraph<NSGVertex, NSGEdge>();
	
	private double minProp, maxProp;
	private SARSet sarset;
	
	private Set<Target> additionalSets = new HashSet<Target>();
	
	
	
	private void init() {
		
		//Get the Property distribution
		minProp = sarset.getCentralPropertyMin();
		maxProp = sarset.getCentralPropertyMax();
		
		
		//If the Range is 0, make it one
		if (minProp == maxProp) {
			minProp -= 0.5;
			maxProp += 0.5;
		}
		
		
		//Initialize SAR Vertices
		for (Molecule m : this.sarset) {
			
			nsGraph.addVertex(
					new NSGVertex(
							m, 
							this.sarset.getLocalScore(m),
							this.sarset.getCentralProperty(m),
							true
					)
			);
		}
		
		
		//Initialize SAR Edges
		for (NSGVertex v1: nsGraph.getVertices()) {
			
			//MoleculeSet ms = sarset.getNeighbors(v1.getMol(), sarset.getCutoff());
			for (NSGVertex v2: nsGraph.getVertices()) {
				
				if (v1.equals(v2))
					continue;
				
				
				
					
				
				
				EdgeType edgeType = 	(this.getCentralProperty(v1) == 	this.getCentralProperty(v2)) ? EdgeType.UNDIRECTED : EdgeType.DIRECTED;			
				
				//if (ms.contains(v2.getMol())) {
				if (sarset.getPrecalcTc(v1.getMol(), v2.getMol()) > sarset.getCutoff()) {	
				
					if (this.getCentralProperty(v1) < this.getCentralProperty(v2)) {
					
					nsGraph.addEdge(
							
							new NSGEdge(v1, v2),
							
							v1,
							v2,
							edgeType
					);
					}
					else {
						nsGraph.addEdge(
								
								new NSGEdge(v2, v1),
								
								v2,
								v1,
								edgeType
						);
					}
				} //if
				
			} //inner loop through neighbors
		} //outer loop through vertices	
	}
	
	
	private void deleteSingleVertices() {
		ArrayList<NSG.NSGVertex> todelete = new ArrayList<NSG.NSGVertex>();
		
		for (NSG.NSGVertex v : nsGraph.getVertices()) {
			if (nsGraph.getNeighbors(v).isEmpty())
				todelete.add(v);
		}
		
		for (NSG.NSGVertex v : todelete)
			nsGraph.removeVertex(v);
		
	}
	
	public NSG (SARSet s) {
		
		this.sarset = s;
				
		init();
		
		//System.err.println("Vertices before singleton filtering: " + nsGraph.getVertexCount());
		//deleteSingleVertices();
		//System.err.println("Vertices after singleton filtering: " + nsGraph.getVertexCount());
		
	}
	
	public String toString () {
		return nsGraph.toString();
	}
	
	
	public Collection<NSGVertex> getSarVertices() {
		
		 Collection<NSGVertex> sv = new HashSet<NSGVertex>();
		 
		 for (NSGVertex v : nsGraph.getVertices()) {
			 if (v.isInSarSet())
				 sv.add(v);
		 }
		  
		 return sv;
		
	}
	
	
	
	
	private void addVerticesOfSet (MoleculeSet ms, Color col, double threshold) {
		
		//Check if the molecule is already present in the graph
		for (NSGVertex v: this.nsGraph.getVertices()) {
			
			if (ms.contains(v.getMol())) {
				v.addToSet(ms);
				v.setSetColor(col);
				v.setThreshold(threshold);
				ms.remove(v.getMol());		
			}
		}
		
		
		Collection<NSGVertex> sarvertices = this.getSarVertices();
		
		
		
		for(Molecule m: ms) {
			NSGVertex vertexCreated = null;
			for (NSGVertex v2: sarvertices) {
				
				Molecule m2 = v2.getMol();
				
				if (m.Tc(m2) > threshold) {
					
					//If not already created, add new Vertex to the graph
					if (vertexCreated == null) {
						vertexCreated = new NSGVertex(m,-1,-1,false);
						vertexCreated.addToSet(ms);	//Add the Vertex to the current set
						vertexCreated.setSetColor(col);
						vertexCreated.setThreshold(threshold);
						this.nsGraph.addVertex(vertexCreated);
					}
					
					//Add edge between new Vertex and molecule in graph
					this.nsGraph.addEdge(new NSGEdge(vertexCreated, v2), vertexCreated, v2, EdgeType.UNDIRECTED);
					
					v2.setHasAddNeighbor(true);
					v2.setSetColor(vertexCreated.getSetColor());
					
				}
			}
		}
		
	}
	
	
	public boolean addSet (Target t, Color col, double threshold) {
		
		if (!additionalSets.add(t))
			return false;
		
		addVerticesOfSet(t.getMoleculeSet(), col, threshold);
		
		return true;
		
	}
	
	public boolean addSet (Target t, Color col) {
		
		if (!additionalSets.add(t))
			return false;
		
		addVerticesOfSet(t.getMoleculeSet(), col, this.sarset.getCutoff());
		
		return true;
		
	}
	
	public boolean contains (Molecule mol) {
		
		
		for (NSGVertex v : nsGraph.getVertices()) {
			if (mol.equals(v.getMol()))
				return true;
		}
		
		return false;
		
	}
	
	public MoleculeSet allMolecules () {
		
		MoleculeSet set = new MoleculeSet();
		
		for (NSGVertex v : this.nsGraph.getVertices()) {
			set.add(v.getMol());
		}
		
		return set;
		
	}
	
	public int getNumberMolecules() {
		
		return nsGraph.getVertexCount();
	}
	
	public Collection<Target> allTargets () {
		Collection<Target> tc = new HashSet<Target>();
		
		tc.addAll(sarset.getAllTargets());
		tc.addAll(additionalSets);
		
		
		return tc;
			
	}
	
	public Collection<Target> allAddTargets () {
		Collection<Target> tc = new HashSet<Target>();
		
		tc.addAll(additionalSets);
		
		return tc;
			
	}
	
	
	public NSGVertex getVertex(Molecule mol) {
		
		for (NSGVertex v: this.nsGraph.getVertices()) {
			if (v.getMol().equals(mol))
				return v;
		}
		
		return null;
	}
	
	public Graph<NSGVertex, NSGEdge> getGraph() {
		
		return this.nsGraph;
	}
	
	public String graphTitle() {
		return this.sarset.toString();
	}
	
	public String getCentralPropName() {
		return sarset.getCentralPropertyName();
	}
	
	public double getCentralProperty(NSG.NSGVertex v) {
		
		if (v.isInSarSet())
			return sarset.getCentralProperty(v.getMol());
		else
			return nsGraph.getNeighbors(v).size() / 100;
	}
	
	public String getCentralPropertyForTable(NSG.NSGVertex v) {
		if (v.isInSarSet())
			return sarset.getCentralPropertyForTable(v.getMol());
		else
			return String.valueOf(nsGraph.degree(v)) + " neighbors";
	}
	
		
	
	public double getLocalScore (NSG.NSGVertex v) {
		if (v.isInSarSet())
			return sarset.getLocalScore(v.getMol());
		else
			return 2;
		
	}
	
	public String getLocalScoreForTable (NSG.NSGVertex v) {
		if (v.isInSarSet())
			return String.format("%.3f", sarset.getLocalScore(v.getMol()));
		else
			return "n/a";
		
	}
	
	public boolean isSSR() {
		return (!sarset.isSAR());
	}
	
	public String[] getGlobalScores () {
		String[] locals = new String[3];
		
		locals[0] = String.format("%.3f", sarset.getCont());
		locals[1] = String.format("%.3f", sarset.getDisc());
		locals[2] = String.format("%.3f", sarset.getSARI());
		
		return locals;
		
	}
	
	public String[] getTragetNames()
	{
		
		ArrayList<Target> allT = sarset.getAllTargets();
		
		allT.addAll(additionalSets);
		
		String allTs[] = new String[allT.size()];
		
		for (int i = 0; i < allT.size(); i++) {
			
			allTs[i] = allT.get(i).toString();
			
		}
		
		return allTs;
		
	}
	
	public double getTc (NSGVertex v1, NSGVertex v2) {
		
		return sarset.getPrecalcTc(v1.getMol(), v2.getMol());
		
	}
	
	public double getDelta(NSGVertex v1, NSGVertex v2) {
		
		return Math.abs(this.getCentralProperty(v1) - this.getCentralProperty(v2));
	}
	
	public double getCutoff() {
		return sarset.getCutoff();
	}
	
	public SARSet getSarSet() {
		return sarset;
	}
	
	
	
	
	

}
