/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * 
 */
package bit.lsi.sar.nsg;

import java.io.IOException;
import java.util.*;


/**
 * @author lounkine
 *
 */
public class Normalizer {

	
	private NSGReader nsgreader;
	
	//Collections that store the raw scores
	private Collection<Double> globalSARICont = new HashSet<Double>();
	private Collection<Double> globalSARIDisc = new HashSet<Double>();
	private Collection<Double> globalSSRICont = new HashSet<Double>();
	private Collection<Double> globalSSRIDisc = new HashSet<Double>();
	private Collection<Double> localSARI = new HashSet<Double>();
	private Collection<Double> localSSRI = new HashSet<Double>();
	
	
	private final static double[] percentiles = {15,30,45,60,80,100};
	private ArrayList<Double> tcValues = new ArrayList<Double>(10000000);
	private double[] tcBins = new double[6]; 
	private double cutoff;
	
	
	
	private HashMap<String, Double> normalizationFactors = new HashMap<String, Double>();
	
	
	
	
	private void calcStatistics() {
		
		
		HashMap<String,Double> readercs = nsgreader.getNormalizationConstants();
		readercs.put("CUTOFF", cutoff);
		
		SARSet.setNorm(readercs);
		SSRSet.setNorm(readercs);
		
		Runtime rt = Runtime.getRuntime();
		
		for (Target t: nsgreader.getSARTargets()) {
			
			rt.gc();
			
			System.err.println("Making SARSet for Target " + t);
			
			SARSet sarset = t.getSARSet();
			
			
			tcValues.addAll(sarset.getAllTcValues(cutoff));
			
			globalSARICont.add(sarset.getRawCont());
			globalSARIDisc.add(sarset.getRawDisc());
			localSARI.addAll(sarset.getRawLocalScores());
			
			
			//all SSR sets with >= 5 molecules
			for (Target t2: nsgreader.getSARTargets()) {
				
				if (t.compareTo(t2) >= 0)
					continue;
			
				
				MoleculeSet curMs = t.getMoleculeSet();
				curMs.retainAll(t2.getMoleculeSet());
					
				if (curMs.size() < 5)
					continue;
									
				
				System.err.println("Making SSRSet for Target " + t + " vs. " + t2);
				sarset = SSRSet.getSSRSet(t, t2);
				
				tcValues.addAll(sarset.getAllTcValues(cutoff));
				
				globalSSRICont.add(sarset.getRawCont());
				globalSSRIDisc.add(sarset.getRawDisc());
				localSSRI.addAll(sarset.getRawLocalScores());
			}
			
			
		}
		
		System.err.println("Finished building SAR / SSR Sets");
		System.err.println("Calculating statistics...");
		
		
		HashMap<String, Double> curStats;
		
		curStats = calcStats(globalSARICont);
		normalizationFactors.put("globalSARIContMean", curStats.get("avg"));
		normalizationFactors.put("globalSARIContSd", curStats.get("sd"));
	
		curStats = calcStats(globalSARIDisc);
		normalizationFactors.put("globalSARIDiscMean", curStats.get("avg"));
		normalizationFactors.put("globalSARIDiscSd", curStats.get("sd"));
		
		curStats = calcStats(globalSSRICont);
		normalizationFactors.put("globalSSRIContMean", curStats.get("avg"));
		normalizationFactors.put("globalSSRIContSd", curStats.get("sd"));
	
		curStats = calcStats(globalSSRIDisc);
		normalizationFactors.put("globalSSRIDiscMean", curStats.get("avg"));
		normalizationFactors.put("globalSSRIDiscSd", curStats.get("sd"));
		
		curStats = calcStats(localSARI);
		normalizationFactors.put("localSARImean", curStats.get("avg"));
		normalizationFactors.put("localSARIsd", curStats.get("sd"));
		
		curStats = calcStats(localSSRI);
		normalizationFactors.put("localSSRImean", curStats.get("avg"));
		normalizationFactors.put("localSSRIsd", curStats.get("sd"));
		
		
		System.err.print("Sorting Tc value vector...");
		Collections.sort(tcValues);
		System.err.println("Done.");
		
		//System.err.println(tcValues);
		
		for (int i = 0; i < 5; i++) {
			
			tcBins[i] = tcValues.get( tcValues.size() - (int) ( percentiles[i] * (double) tcValues.size()/100) );
					
		}
		
		tcBins[5] = 0;
		
		
		
		
	}
	
	
	public static HashMap<String,Double> calcStats (Collection<Double> c) {
		
		HashMap<String,Double> stats = new HashMap<String,Double>();
		
		double sum = 0;
		for (Double v : c)
			sum += v;
		
		double avg = sum / c.size();
		
		double varsum = 0;
		
		for (Double v : c)
			varsum += Math.pow((v - avg), 2);
		
		varsum /= (c.size() - 1);
		
		double sd = Math.sqrt(varsum);
		
		stats.put("avg", avg);
		stats.put("sd", sd);
		
		
		return stats;
		
		
	}
	
	
public static HashMap<String,Double> calcExtendedStats (Collection<Double> c) {
		
		HashMap<String,Double> stats = new HashMap<String,Double>();
		ArrayList<Double> ac = new ArrayList<Double>(c.size());
		
		double sum = 0;
		
		for (Double v : c) {
			ac.add(v);
			sum += v;
		}
			
			
		
		double avg = sum / c.size();
		
		double varsum = 0;
		
		for (Double v : c)
			varsum += Math.pow((v - avg), 2);
		
		varsum /= (c.size() - 1);
		
		double sd = Math.sqrt(varsum);
		
		stats.put("avg", avg);
		stats.put("sd", sd);
		stats.put("min", Collections.min(c));
		stats.put("max", Collections.max(c));
				
		Collections.sort(ac);
		
		
		if (ac.size() % 2 == 0)
			stats.put("median", ac.get(ac.size() / 2));
		else
			stats.put("median", (
					(ac.get((int) ac.size() / 2))
					+ (ac.get(1 + (int) ac.size() / 2))
					) / 2
					);
		
		stats.put("q25", (ac.get((int) (ac.size() * 0.25))));
		stats.put("q75", (ac.get((int) (ac.size() * 0.75))));
		
			
		return stats;
		
		
	}
	
	public Normalizer(String setFn) throws IOException {
		
		nsgreader = new NSGReader(setFn);
		cutoff = nsgreader.getCutoff();	
		calcStatistics();
			
	}
	
	public Normalizer(String setFn, double cutoff) throws IOException {
		
		nsgreader = new NSGReader(setFn);
		this.cutoff = cutoff;	
		calcStatistics();
			
	}
	
	
	public String getNormalizationFactors() {
		
		String s = "";
		
		for (String key : normalizationFactors.keySet()) {
			
			s += key + "," + normalizationFactors.get(key) + "\n";
			
		}
		
		s += "CUTOFF," + this.cutoff + "\n";
		
		s += "RADIALS," + tcBins[0];
			
		for (int i = 1; i < tcBins.length; i++)
			s += " " + tcBins[i];
		
		return s;
	}
	
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		Normalizer norm = null;
		
		if (args.length == 2) {
			
			try {
				norm = new Normalizer(args[0], Double.parseDouble(args[1]));
			} catch (NumberFormatException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			
		}
		else {
			
			try {
				norm = new Normalizer(args[0]);
			} 
			catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
		
		if (norm != null) {
			
			System.out.print(norm.getNormalizationFactors());
			
		}
		 

	}

}
