/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * 
 */
package bit.lsi.sar.nsg;




/**
 * 
 * Potency Object holding the potency in Nanomolar.
 *  
 * @author lounkine
 *
 */
public class Potency {
	
	
	/**
	 * 
	 */
	//private static final long serialVersionUID = -7551358477418679781L;

	private static String[] units = {"M", "mM", "uM", "nM", "pM", "fM"};
	
	private double nM;
	private String potstring;
	
	/** Potency Units from Femtomolar to Molar. Micromlar is uM.*/
	public static enum Units {fM, pM, nM, uM, mM, M};
	
	
	/** @param nM The potency in Nanomolar*/
	public Potency (double nM) {
		
		this.nM = nM;
		this.potstring = calcPotstring();
	}
	
	/** @param value potency
	 *  @param unit The unit of value*/
	public Potency (double value, Potency.Units unit) {
		
		double factor = 1;
		
		switch (unit) {
		
			case M:		factor = 1000000000; break;
			case mM: 	factor = 1000000; break;
			case uM: 	factor = 1000; break;
			case nM:	factor = 1; break;
			case pM:	factor = 0.001; break;
			case fM:	factor = 0.000001; break;
			default: factor = 1;
		}
		
		
		this.nM = value * factor;
		this.potstring = calcPotstring();
	}
	
	
	/** Potency with appropriate Unit. uM is Micromolar.*/
	public String toString () {
		
		return this.potstring;
		
	}
	
	
	public static String potString (double pKi) {
		
		double molar = Math.pow(10, -pKi);
		
		int i = 0;
		while (molar < 1) {
			molar *= 1000;
			i++;
		}
		
		return String.format("%.3f", molar) + " " + units[i];
	}
	 
	
	private String calcPotstring () {
		
		return potString(this.getpKi());
	}
	
	
	/**Potency in nM*/
	public double getPot() {
		return this.nM;
	}
	
	/**Potency as pKi / pIC50*/
	public double getpKi() {
		return 9 - Math.log10(this.nM);
	}
	
	/** Calculates the logarithm of the selectivity ratio of two potencies.
	 * Negative values mean selective for pot1.
	 * Positive values mean selective for pot2.
	 * Non-selective returns a value of 0*/
	public static double pSR (Potency pot1, Potency pot2) {
		
		return Math.log10( pot1.getPot() / pot2.getPot() );
		
	}
	
	/** Converts a pSR value to text of the form 1:x or x:1*/
	public static String pSRtext (double pSR) {
		
		
		String spSR = String.format("%.2f", Math.pow(10, Math.abs(pSR)));
		
		
		if (pSR == 0) {
			return "1:1";
		}
		else if (pSR < 0) {
			return spSR + ":1";
		}
		else  {
			return "1:" + spSR;
		}
		
	}
	
	/** Calculates the logarithm of the selectivity ratio of this and another potency.
	 * Negative values mean selective for this.
	 * Positive values mean selective for pot2.
	 * Non-selective returns a value of 0*/
	public double pSR (Potency pot2) {
	
		//System.err.println(pot2);
		return Math.log10( nM / pot2.getPot() );
	
	}	

}
