/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package bit.lsi.sar.nsg;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;


import cern.jet.stat.Probability;


public class SARSet extends MoleculeSet {

	/**
	 * 
	 */
	private static final long serialVersionUID = 5888801669098845203L;
	
	
	
	private Target t1;
	
	protected HashMap<Molecule,Double> localScores = new HashMap<Molecule,Double>();
	protected HashMap<Molecule,Double> rawlocalScores;
	
	protected double rawCont = 0;
	protected double rawDisc = 0;
	
	protected double cont = 0;
	protected double disc = 0;
	protected double sari = 0;
	
	private static double cutOff, gCMean, gCVAR, gDMean, gDVAR, lMean, lVAR;
	
	
	
	
	public static void setNorm(HashMap<String,Double> norms) {
		
		cutOff = norms.get("CUTOFF");
		
		gCMean = norms.get("globalSARIContMean");
		gCVAR =  Math.pow( norms.get("globalSARIContSd"), 2 );
		
		gDMean = norms.get("globalSARIDiscMean");
		gDVAR =  Math.pow( norms.get("globalSARIDiscSd"), 2 );
		
		lMean = norms.get("localSARImean");
		lVAR =  Math.pow( norms.get("localSARIsd"), 2 );	
	}
	
	public static double getStaticCutoff() {
		return cutOff;
	}
	
	
	
	
	private void normalizeScores() {
		
		
		cont = Probability.normal(gCMean, gCVAR, rawCont);
		disc = Probability.normal(gDMean, gDVAR, rawDisc);
		sari = 0.5 * (cont + (1-disc));
		
		rawlocalScores = new HashMap<Molecule,Double>(localScores);
		
		for (Molecule key : localScores.keySet()) {
			
			localScores.put(key, Probability.normal(lMean, lVAR, localScores.get(key)));
			
		}
		
	}
	
	public Collection<Double> getRawLocalScores() {
		
		Collection<Double> lss = new HashSet<Double>(rawlocalScores.values());
		
		
		return lss;
	}
	
	
	
	
	private void calcRawCont (){
		
		double weightSum = 0;
		
		
		//Nested Molecule loop that compares all the molecules
		for (Molecule m1 : this) {
			for (Molecule m2 : this) {
				
				if (m1.getInternalId() >= m2.getInternalId())
					continue;
				
				//Calculate the weight
				double curWeight = 
					( m1.getPotency(t1).getpKi() * m2.getPotency(t1).getpKi() )
					/ (1 + Math.abs( m1.getPotency(t1).getpKi() - m2.getPotency(t1).getpKi() )
				);
				
				weightSum += curWeight;
				
				double curSimTerm = 1 / 
					(1 + this.getPrecalcTc(m1, m2));
				
				rawCont += curWeight * curSimTerm;
			}
		}
		
		rawCont /= weightSum;
		
	}
	
	
	private void calcRawDisc() {
		
		int totalPairs = 0;
		
		for (Molecule m1 : this) {
			for (Molecule m2 : this) {
				
				double potDiff = Math.abs(
						m1.getPotency(t1).getpKi() 
					  - m2.getPotency(t1).getpKi()
				);
				
				if (m1.getInternalId() >= m2.getInternalId())
					continue;
				
				double sim = this.getPrecalcTc(m1, m2);
				
				//The conditions to calculate the globals discontinuity score...
				if (
						(sim <= cutOff)
					||	(potDiff <= 1)
						
				)
					continue;
				
				//Increase the normalization factor
				totalPairs++;
				
				//Calculate the Term
				double curTerm = potDiff * sim;
				
				rawDisc += curTerm;
			}
		}
		
		if (totalPairs > 0)
			rawDisc /= totalPairs;
		
	}
	
	
	private void calcLocalDisc () {
		
		
		
		for (Molecule m1 : this) {
			
			int totalPairs = 0;
			double curLocalScore = 0;
			
			//System.err.println(m1);
				
			for (Molecule m2 : this) {
				
				double sim = this.getPrecalcTc(m1, m2);
				if (sim <= cutOff)
					continue;
				
				double potDiff = Math.abs(
						m1.getPotency(t1).getpKi() 
					  - m2.getPotency(t1).getpKi()
				);
				
				totalPairs++;
				
				
					curLocalScore += potDiff * this.getPrecalcTc(m1, m2);
				
			} //Neighbors of m1
			
			if (totalPairs > 0)
				curLocalScore /= totalPairs;
			
			localScores.put(m1, curLocalScore);					
		
		} //all molecules
		
	}
	
	
	private void init() {
	
		System.err.println("SARSet: Initializing...");
		System.err.println("SARSet: Calculating Tc similarity matrix...");
		
		
		this.precalcTcMatrix();
		
		System.err.println("SARSet: Tc similarity matrix calculated");
		
		
		calcRawCont();
		System.err.println("SARSet: Calculated raw continuity scores.");
		
		calcRawDisc();
		System.err.println("SARSet: Calculated raw discontinuity scores.");
		calcLocalDisc();
		System.err.println("SARSet: Calculated raw local scores.");
		normalizeScores();
		System.err.println("SARSet: Normalized scores.");
		
		
		
		
	}
	
	public SARSet(Collection<Molecule> c, Target t) {
		super(c);
		
		this.t1 = t;
		
		
		System.err.println(this.size() + "Molecules added to SARSet.");
		
		init();
		
	}
	
	protected SARSet(Collection<Molecule> c) {
		super(c);
	}
	
	public static MoleculeSet getSARSet( Target t) {
		
		return t.getSARSet();
		
	}
	
	public double getRawDisc() {
		return this.rawDisc;
	}
	
	public double getRawCont() {
		return this.rawCont;
	}
	
	
	public double getCont() {
		
		return this.cont;	
		
	}
	
	public double getDisc () {
		return this.disc;
	}
	
	public double getSARI () {
		return this.sari;
	}
	
	public double getLocalScore(Molecule m) {
		return localScores.get(m);
	}
	
	public double getLocalScoreMin() {
		
		return Collections.min(localScores.values());
		
	}
	
	public double getLocalScoreMax() {
		
		return Collections.max(localScores.values());
		
	}
	
	
	/** Returns the Property that is later used to generate color*/
	public double getCentralProperty(Molecule m){
		return m.getPotency(t1).getpKi();
	}
	
	public String getCentralPropertyName() {
		return "Potency";
	}
	
	public boolean isSAR() {
		return true;
	}
	
	public String getCentralPropertyForTable(Molecule m) {
		if (this.contains(m))
			return m.getPotency(t1).toString();
		else
			return "Additional Set";
	}
	
	public double getCentralPropertyMin() {
	
		ArrayList<Double> scores= new ArrayList<Double>(this.size());
		
		
		for(Molecule m : this) {
			scores.add(this.getCentralProperty(m));
		}
		
		return Collections.min(scores);
	}
	
	public double getCentralPropertyMax() {
		
		ArrayList<Double> scores= new ArrayList<Double>(this.size());
		
		
		for(Molecule m : this) {
			scores.add(this.getCentralProperty(m));
		}
		
		return Collections.max(scores);
		
	}
	
	public double getCutoff() {
		return cutOff;
	}
	
	public String toString (){
		return this.t1.toString();
	}
	
	public ArrayList<Target> getAllTargets() {
		 ArrayList<Target> tc = new ArrayList<Target>();
		 tc.add(t1);
		 
		 return tc;
	}
	
	
	public int getNumberMolecules() {
		
		return this.size();
	}
	
	
}
