/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


package bit.lsi.sar.nsg;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.commons.collections15.Transformer;

import bit.lsi.sar.nsg.NSG.NSGEdge;
import bit.lsi.sar.nsg.NSG.NSGVertex;
import cern.jet.stat.Probability;
import edu.uci.ics.jung.algorithms.filters.Filter;
import edu.uci.ics.jung.algorithms.filters.FilterUtils;
import edu.uci.ics.jung.algorithms.filters.KNeighborhoodFilter;
import edu.uci.ics.jung.algorithms.shortestpath.DijkstraShortestPath;
import edu.uci.ics.jung.graph.DelegateForest;
import edu.uci.ics.jung.graph.DirectedSparseGraph;
import edu.uci.ics.jung.graph.Forest;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.SparseGraph;
import edu.uci.ics.jung.graph.util.EdgeType;

public class SARTree {
	
	private DirectedSparseGraph<NSGVertex,NSGEdge> graph;
	private DelegateForest<NSGVertex,NSGEdge> tree;
	private NSG nsg;
	private PathContainer allPaths = new PathContainer();
	
	
	
	class PathContainer extends HashSet<Path> {
		
			
		/**
		 * 
		 */
		private static final long serialVersionUID = -2789786499641787137L;
		private double mean = 0;
		private double var = 1;
		
		
		public void calcStats() {
			
			Collection<Double>  allPathScores = new ArrayList<Double>();
			
			for (Path p: this)
				allPathScores.add(p.getPathScore());
					
			HashMap<String,Double> scoreStats = Normalizer.calcStats(allPathScores);
			mean = scoreStats.get("avg");
			var = Math.pow(scoreStats.get("sd"), 2);
			
			
		}
		
		
		public Path getMaxPath(NSG.NSGVertex v) {
			
			double maxScore = 0;
			Path maxPath = null;
			
			for (Path p : this) {
				
				if (p.getVertices().contains(v) && p.getPathScore() > maxScore) {
					maxPath = p;
					maxScore = p.getPathScore();
				}
				
			}
			
			return maxPath;
		}
		
		public double maxEdgeScore (NSG.NSGEdge e) {
			
			double maxScore = 0;
		
			for (Path p : this) {
				
				if (p.contains(e) && p.getPathScore() > maxScore) {
					maxScore = p.getPathScore();
				}
				
			}
			
			return maxScore;
					
		}
		
		public double maxEdgeWeight(NSG.NSGEdge e) {
			
			return Probability.normal(mean, var, maxEdgeScore(e));
			
		}
		
	}
	
	
	class Path extends LinkedList<NSG.NSGEdge> {
		
		/**
		 * 
		 */
		private static final long serialVersionUID = 5082263650102576302L;
		private double pathCost = 0;
		private double pathScore = 0;
		
		private double flPotDiff;
		
		private HashSet<NSG.NSGVertex> vertices;
		
		
		private void init() {
			
			vertices = new HashSet<NSG.NSGVertex>(this.size() + 1);
			
			NSG.NSGEdge firstToLastEdge = nsg.new NSGEdge(this.getFirst().getVertices()[0], this.getLast().getVertices()[1]);
			flPotDiff = firstToLastEdge.getCentralPropDiff();
			
			vertices.add(firstToLastEdge.getVertices()[0]);
			
			double curPathScore = 0;
			for (NSG.NSGEdge e : this) {
				
				vertices.add(e.getVertices()[1]);
				
				pathCost += e.getFancy();
				curPathScore += Math.abs((flPotDiff / this.size()) -  e.getCentralPropDiff());
			}
			
			pathScore = (this.size() - 1) * flPotDiff / (1+curPathScore);
			
		}
		
		public Path(List<NSG.NSGEdge> edges) {
			super();
			this.addAll(edges);
			
			init();
		}
		
		
		public double getPathScore() {
			return pathScore;
		}
		
		public Set<NSG.NSGVertex> getVertices() {
			return vertices;
		}
		
	}
	
	
	
	private boolean initgraph(NSG nsg, NSG.NSGVertex v) {
		
		
		this.nsg = nsg;
		
		Filter<NSGVertex,NSGEdge> reachableInFilter = new KNeighborhoodFilter<NSGVertex, NSGEdge>(v, 15, KNeighborhoodFilter.EdgeType.IN);
		Filter<NSGVertex,NSGEdge> reachableOutFilter = new KNeighborhoodFilter<NSGVertex, NSGEdge>(v, 15, KNeighborhoodFilter.EdgeType.OUT);
		
		
		
		Graph<NSGVertex,NSGEdge> ngi = reachableInFilter.transform(nsg.getGraph());
				
		Graph<NSGVertex,NSGEdge> ngo = reachableOutFilter.transform(nsg.getGraph());
		
		Graph<NSG.NSGVertex,NSG.NSGEdge> ng = new SparseGraph<NSGVertex, NSGEdge>();
				
		Collection<NSG.NSGVertex> vc = 	new ArrayList<NSGVertex>(ngo.getVertexCount() + ngi.getVertexCount());
		
		vc.addAll(ngo.getVertices());
		vc.addAll(ngi.getVertices());
		
		vc.retainAll(nsg.getSarVertices());
		
		ng = FilterUtils.createInducedSubgraph(vc, nsg.getGraph());
		
		
		Collection<NSG.NSGEdge> redges = new HashSet<NSG.NSGEdge>();
		
		for (NSG.NSGEdge e : ng.getEdges()) {
			if (e.hasZeroDiff())
				redges.add(e);
		}
		
		for (NSG.NSGEdge e : redges)
			ng.removeEdge(e);
		
		
		
		System.err.println("NG : " + ng.getVertexCount());
		
		DijkstraShortestPath<NSGVertex,NSGEdge> sp = new DijkstraShortestPath<NSGVertex,NSGEdge>(ng, new Transformer<NSGEdge, Number>() {

			@Override
			public Number transform(NSGEdge e) {
				return e.getFancy();
			}
			
		});
		
		final ArrayList<NSGEdge> pathEdges = new ArrayList<NSGEdge>();
		
		List<NSG.NSGEdge> curEdges;
		
		for (NSGVertex hv : ng.getVertices()) {
			
			if (hv.equals(v))
				continue;
			
			if (ng.getPredecessorCount(hv) == 0 ) {
				curEdges = sp.getPath(hv, v);
			}
			else if (ng.getSuccessorCount(hv) == 0) {
				
				curEdges = sp.getPath(v, hv);
			}
			else
				continue;
			
			
			if (curEdges.isEmpty())
				continue;
			
			allPaths.add(new Path(curEdges));
			pathEdges.addAll(curEdges);
			
		}
		
		//Normalize over all Path costs (using normal distribution)
		allPaths.calcStats();
		
		
		for (NSG.NSGEdge e : ng.getEdges()) {
			if (!pathEdges.contains(e))
				ng.removeEdge(e);
		}
		
		
		graph = new DirectedSparseGraph<NSGVertex, NSGEdge>();
		for (NSG.NSGVertex vh : ng.getVertices()) {
			if (ng.degree(vh) > 0)
				graph.addVertex(vh);
		}
		for (NSG.NSGEdge e : ng.getEdges()) {
			graph.addEdge(e, ng.getEndpoints(e), EdgeType.DIRECTED);
		}
		
		tree = new DelegateForest<NSGVertex, NSGEdge>(graph);
				
		return true;
		
	}
	
	public SARTree (NSG nsg, NSG.NSGVertex v) {
		
			
			initgraph(nsg, v);
			
			//System.err.println(graph);
			
	}
	
	public double getEdgePathWeight(NSG.NSGEdge e) {
		return allPaths.maxEdgeWeight(e);
	
	}
	
	public String getEdgePathWeightString(NSG.NSGEdge e) {
		return "Score: " + String.format("%.3f", allPaths.maxEdgeScore(e));
		
	}
	
	public Graph<NSG.NSGVertex, NSG.NSGEdge> getGraph(){
		return graph;
	}
	
	public Collection<NSG.NSGVertex> getPathVertices (NSG.NSGVertex v) {
		
		return allPaths.getMaxPath(v).getVertices();
		
	}
	
	public Forest<NSG.NSGVertex, NSG.NSGEdge> getTree() {
		return tree;
	}
	
	public NSG getNSG() {
		return nsg;
	}
	
		
	

}
