/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package bit.lsi.sar.nsg;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;

import cern.jet.stat.Probability;


public class SSRSet extends SARSet {

	/**
	 * 
	 */
	private static final long serialVersionUID = 712982970839617315L;
	
	private Target t1, t2;
	
	//private HashMap<Molecule,Double> localScores = new HashMap<Molecule,Double>();
	
	private static double cutOff, gCMean, gCVAR, gDMean, gDVAR, lMean, lVAR;




	public static void setNorm(HashMap<String,Double> norms) {
	
		cutOff = norms.get("CUTOFF");
	
		gCMean = norms.get("globalSSRIContMean");
		gCVAR =  Math.pow( norms.get("globalSSRIContSd"), 2 );
	
		gDMean = norms.get("globalSSRIDiscMean");
		gDVAR =  Math.pow( norms.get("globalSSRIDiscSd"), 2 );
	
		lMean = norms.get("localSSRImean");
		lVAR =  Math.pow( norms.get("localSSRIsd"), 2 );	
	}
	
	
	
	
	
	private void normalizeScores() {
		
		cont = Probability.normal(gCMean, gCVAR, rawCont);
		disc = Probability.normal(gDMean, gDVAR, rawDisc);
		sari = 0.5 * (cont + (1-disc));
		
		rawlocalScores = new HashMap<Molecule,Double>(localScores);
		//Normalize local scores
		/*
		Collection<Double> localScoreValues = localScores.values();
		for (double curVal : localScoreValues) {
			curVal = Probability.normal(lMean, lVAR, curVal);
		}*/
		
		for (Molecule key : localScores.keySet()) {
			
			localScores.put(key, Probability.normal(lMean, lVAR, localScores.get(key)));
			
		}
		
	}
	
	
	private void calcRawCont (){
		
		double weightSum = 0;
		
		
		//Nested Molecule loop that compares all the molecules
		for (Molecule m1 : this) {
			for (Molecule m2 : this) {
				
				if (m1.getInternalId() >= m2.getInternalId())
					continue;
				
				//get the pSR values
				double pSR1 = m1.getpSR(this.t1, this.t2);
				double pSR2 = m2.getpSR(this.t1, this.t2);
				
				//Calculate the weight
				double curWeight = 
					( pSR1 * pSR2 ) / (1 + Math.abs( pSR1 - pSR2 )
				);
				
				weightSum += curWeight;
				
				double curSimTerm = 1 / 
					(1 + this.getPrecalcTc(m1, m2));
				
				rawCont += curWeight * curSimTerm;
			}
		}
		
		rawCont /= weightSum;
		
	}
	
	
	private void calcRawDisc() {
		
		int totalPairs = 0;
		
		for (Molecule m1 : this) {
			for (Molecule m2 : this) {
				
				double srDiff = Math.abs(
						m1.getpSR(this.t1, this.t2) 
					  - m2.getpSR(this.t1, this.t2)
				);
				
			
				if (m1.getInternalId() >= m2.getInternalId())
					continue;
				
				double sim = this.getPrecalcTc(m1, m2);
				//The conditions to calculate the globals discontinuity score...
				if (
						(sim <= cutOff)
					||	(srDiff <= 1)
						
				)
					continue;
				
				
				//Increase the normalization factor
				totalPairs++;
				
				//Calculate the Term
				double curTerm = srDiff * sim;
				
				rawDisc += curTerm;
			}
		}
		
		if (totalPairs > 0)
			rawDisc /= totalPairs;
		
	}
	
	
	private void calcLocalDisc () {
		
		
		
		for (Molecule m1 : this) {
			
			int totalPairs = 0;
			double curLocalScore = 0;
			
			//System.err.println(m1);
			//MoleculeSet curNeighbors = this.getNeighbors(m1, cutOff);
			
			for (Molecule m2 : this) {
				
				double sim = this.getPrecalcTc(m1, m2);
				if (sim <= cutOff)
					continue;
				
				double srDiff = Math.abs(
						m1.getpSR(this.t1, this.t2) 
					  - m2.getpSR(this.t1, this.t2)
				);
				
				totalPairs++;
				
				curLocalScore += srDiff * sim;
			
			} //Neighbors of m1
			
			if (totalPairs > 0)
				curLocalScore /= totalPairs;
			
			localScores.put(m1, curLocalScore);					
		
		} //all molecules
		
	}
	
	
	private void init() {
		
		System.err.println("SSRSet: Initializing...");
		System.err.println("SSRSet: Calculating Tc similarity matrix...");
		
		
		this.precalcTcMatrix();
		
		System.err.println("SSRSet: Tc similarity matrix calculated");
		
		
		calcRawCont();
		System.err.println("SSRSet: Calculated raw continuity scores.");
		
		calcRawDisc();
		System.err.println("SSRSet: Calculated raw discontinuity scores.");
		calcLocalDisc();
		System.err.println("SSRSet: Calculated raw local scores.");
		normalizeScores();
		System.err.println("SSRSet: Normalized scores.");
		
		
		
		
	}
	
	
	
	
	public SSRSet(Collection<Molecule> c, Target t1, Target t2) {
		
		super(c);
		this.t1 = t1;
		this.t2 = t2;
		
		init();
		
	}
	
	
	/** Returns the Property that is later used to generate color*/
	public double getCentralProperty(Molecule m){
		return m.getpSR(this.t1, this.t2);
	}
	
	public String getCentralPropertyName() {
		return "Selectivity";
	}
	
	public String getCentralPropertyForTable(Molecule m) {
		if (this.contains(m))
			return Potency.pSRtext(m.getpSR(this.t1, this.t2));
		else
			return "Additional Set";
	}
	
	
	public double getCentralPropertyMin() {
		
		ArrayList<Double> scores= new ArrayList<Double>(this.size());
		
		
		for(Molecule m : this) {
			scores.add(this.getCentralProperty(m));
		}
		
		return -Math.max(Math.abs(Collections.min(scores)), Math.abs(Collections.max(scores)));
	}
	
	public double getCentralPropertyMax() {
		
		return -this.getCentralPropertyMin();
		
	}
	
	public static SSRSet getSSRSet( Target t1, Target t2) {
		
		MoleculeSet m1 = t1.getMoleculeSet();
		MoleculeSet m2 = t2.getMoleculeSet();
		
		if (m1 == null || m2 == null)
			return null;
		
		
		m1.retainAll(m2);
		
		
		
		if (m1.size() == 0)
			return null;
		
		
		
		return new SSRSet(m1, t1, t2);
		
	}
	
	
	
	public static HashSet<Target> allSSRSetCoTargets(int minMols, Target t, Collection<Target> targets) {
		
		
		HashSet<Target> possibleTargets = new HashSet<Target>();
		
		for (Target curT : targets) {
			
			if (!curT.hasPotencyAnnotation())
				continue;
			
			Collection<Molecule> molcol = curT.sharedMols(t);
			
			if (molcol.size() < minMols)
				continue;
			
			possibleTargets.add(curT);
			
		}
		
		return possibleTargets;
		
			
	}
	
	public boolean isSAR() {
		return false;
	}
	
	public double getCutoff() {
		return cutOff;
	}
	
	public String toString (){
		return this.t1.toString() + " vs. " + this.t2.toString();
	}
	
	public ArrayList<Target> getAllTargets() {
		ArrayList<Target> tc = new ArrayList<Target>();
		 tc.add(t1);
		 tc.add(t2);
		 
		 return tc;
	}
	
	
	
}
