/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package bit.lsi.sar.nsg.swing;

import java.util.Collection;
import java.util.LinkedList;

import javax.swing.table.AbstractTableModel;

import bit.lsi.sar.nsg.Molecule;
import bit.lsi.sar.nsg.NSG;
import bit.lsi.sar.nsg.Target;

public class MoleculeTableModel extends AbstractTableModel {
	
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private final String[] tableColumnNames;
	private final NSG nsg;
	private final LinkedList<NSG.NSGVertex> tableRows;
	
	
	public MoleculeTableModel(String[] columnNames, NSG nsg, LinkedList<NSG.NSGVertex> rows) {
		super();
		this.tableColumnNames = columnNames;
		this.nsg = nsg;
		this.tableRows = rows;
		
		
	}
	
	@Override
	public int getColumnCount() {
		return tableColumnNames.length;
	}

	@Override
	public int getRowCount() {
		
		return tableRows.size();
	}

	@Override
	public Object getValueAt(int row, int col) {
		
		
		switch (col) {
		
		case 0: 
			return new TableEntry(
					tableRows.get(row).getMol().getExternalId()
			);
			
		case 1: 
			return new TableEntry(
					nsg.getCentralProperty(tableRows.get(row)),
					nsg.getCentralPropertyForTable(tableRows.get(row))
			);
			
		case 2: return new TableEntry(
				nsg.getLocalScore(tableRows.get(row)),
				nsg.getLocalScoreForTable(tableRows.get(row))
				);
		
		case 3: 
			
			if (row != 0)
			{
			
				double tc =  nsg.getTc(tableRows.get(row), tableRows.getFirst());
				
				return new TableEntry(tc,				
						String.format("%.2f", tc)
					);
			}
			else
				return new TableEntry(1, "Central");
		
		case 4: //Delta
			
			if (row != 0)
			{
			
				double tc =  nsg.getTc(tableRows.get(row), tableRows.getFirst());
				
				double delta = nsg.getDelta(tableRows.get(row), tableRows.getFirst()); 
				delta *= Math.pow((1 + tc), 2);
				
				return new TableEntry(delta,				
						String.format("%.2f", delta)
					);
			}
			else
				return new TableEntry(Double.POSITIVE_INFINITY, "Central");
		
		
		case 5: //Sets
			Collection<Target> mtargets = tableRows.get(row).getMol().getTargets();
			
			mtargets.retainAll(nsg.allAddTargets());
			
			
			return new TableEntry(mtargets.toString());
		
		
		default: 
			
			String userDesc = tableRows.get(row).getMol().getDescriptorValue(tableColumnNames[col]);
			
			double val;
			try {
				val = Double.parseDouble(userDesc);
			}
			catch (NumberFormatException e) {
				val = 0;
			}
			
			return new TableEntry( val, userDesc);
		
			
	
	}
		
	}
	
	@Override
	public boolean isCellEditable(int row, int col)
    { return false; }
	
	@Override
	public void setValueAt(Object value, int row, int col) {
        // Do nothing
    }
	
	@Override
	public String getColumnName(int column) {
		return tableColumnNames[column];
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public Class getColumnClass(int i) {
		return TableEntry.class;
		
	}
	
	
	/**
	 * 
	 * @param col
	 * @param min
	 * @param max
	 * @return number of items so that min <= value < max
	 */
	public int itemsInBin (int col, double min, double max) {
		
		int num = 0;
		double curV;
		for (int i = 0; i < this.getRowCount(); i++) {
			curV = ((TableEntry) this.getValueAt(i, col)).getValue();
			if (curV >= min && curV < max)
				num++;
			
		}
		
		return num;
		
	}
	
	public Molecule getMoleculeAt(int row) {
		return tableRows.get(row).getMol();
	}

}
