/*
 * This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


package bit.lsi.sar.nsg.swing;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.collections15.Factory;
import org.apache.commons.collections15.map.LazyMap;

import cern.colt.matrix.DoubleFactory2D;
import cern.colt.matrix.DoubleMatrix2D;

import bit.lsi.sar.nsg.Molecule;
import bit.lsi.sar.nsg.NSG;
import edu.uci.ics.jung.algorithms.layout.AbstractLayout;
import edu.uci.ics.jung.algorithms.layout.util.RandomLocationTransformer;
import edu.uci.ics.jung.algorithms.util.IterativeContext;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.Pair;

public class MyFRLayout extends AbstractLayout<NSG.NSGVertex, NSG.NSGEdge>
		implements IterativeContext {

	
	private int maxIterations = 3000;
	private int currentIteration;
	private double temperature;
	private double forceConstant;
	
	private double attraction_multiplier = 1;
	private double repulsion_multiplier = 1;
	
	private double repulseRadius;
	private boolean checked = false;
	private double maxDelta = 5;
	private double coolExp = 1.2;
	
	private HashMap<NSG.NSGVertex, Integer> vertexIds;
	private DoubleMatrix2D dxdy = null;
	
	private boolean mute = false;
	
	
	 
	
	public MyFRLayout(Graph<NSG.NSGVertex, NSG.NSGEdge> g, Dimension d) {
		super(g, new RandomLocationTransformer<NSG.NSGVertex>(d), d);
        //max_dimension = Math.max(d.height, d.width);
        initialize();
	}

	
	public void setCoolExp(double e) {
		this.coolExp = e;
	}
	
	
	public void abort() {
		mute = true;
	}
	
	@Override
	public boolean done() {
		
		if (mute)
			return true;
		if (currentIteration > maxIterations || 1/ temperature < 1 / repulseRadius) { 
            
			if (!checked)
               checked = true;
            
            
			return true; 
        
		} 
        return false;
	}

	@Override
	public void step() {
		
		currentIteration++;
		
		temperature = maxDelta * Math.pow((maxIterations - currentIteration) / (double) maxIterations, coolExp);
		
		//System.err.println("Current Temperature = " + temperature);
		
		//Clear deltas
			
		double xd;
		double yd;
		double ded;
		double rf, af;
		
		int j, k;
		Point2D p1, p2;
		
		 while(true) {
	       try {
		
			//Repulsive force
			for (NSG.NSGVertex v1 : vertexIds.keySet()) {
				for (NSG.NSGVertex v2 : vertexIds.keySet()) {
					
					j = vertexIds.get(v1);
					k = vertexIds.get(v2);
					
					if (j >= k)
						continue;
					
					
					//Obtain difference vector
					p1 = transform(v1);
					p2 = transform(v2);
					
					xd = p1.getX() - p2.getX();
					yd = p1.getY() - p2.getY();
					
					ded = Math.sqrt(xd*xd+yd*yd);
					
					xd /= ded;
					yd /= ded;
					
					//Calculate repulsive "force"
					rf = repulsion_multiplier * forceConstant * forceConstant * (1d /ded - ded*ded/repulseRadius);
					
					//Add to the position change vector
					dxdy.setQuick(j, 0, dxdy.getQuick(j, 0) + xd * rf);
					dxdy.setQuick(k, 0, dxdy.getQuick(k, 0) - xd * rf);
					dxdy.setQuick(j, 1, dxdy.getQuick(j, 1) + yd * rf);
					dxdy.setQuick(k, 1, dxdy.getQuick(k, 1) - yd * rf);
					
					
					
				}
				
			}
			break;
	    }
	       catch(ConcurrentModificationException cme) {}
		 }
	      
		
		//Attractive force
		
		double w;
		
		 while(true) {
		       try {
		
				for (NSG.NSGEdge e: getGraph().getEdges()) {
					
					w = e.getSim() * 10;
					
					Pair<NSG.NSGVertex> pair = getGraph().getEndpoints(e);
					
					p1 = transform(pair.getFirst());
					p2 = transform(pair.getSecond());
					
					j = vertexIds.get(pair.getFirst());
					k = vertexIds.get(pair.getSecond());
					
					xd = p1.getX() - p2.getX();
					yd = p1.getY() - p2.getY();
					ded = Math.sqrt(xd*xd+yd*yd);
					
					if (ded != 0) {
						xd/=ded;
						yd/=ded;
					}
					
					af = attraction_multiplier * ded * ded / forceConstant * w;
					
					//Add to the position change vector
					dxdy.setQuick(j, 0, dxdy.getQuick(j, 0) - xd * af);
					dxdy.setQuick(k, 0, dxdy.getQuick(k, 0) + xd * af);
					dxdy.setQuick(j, 1, dxdy.getQuick(j, 1) - yd * af);
					dxdy.setQuick(k, 1, dxdy.getQuick(k, 1) + yd * af);
					
				}
		
					break;
		       }
		       catch (ConcurrentModificationException cme) {}
		
		 }
		
		       
		//Dampen motion, if needed, and move the points
		while(true) {
			try{
		
		 
		 for (NSG.NSGVertex v : vertexIds.keySet()) {
		
			j = vertexIds.get(v);
			
			ded = Math.sqrt(dxdy.getQuick(j, 0) * dxdy.getQuick(j, 0) + dxdy.getQuick(j, 1) * dxdy.getQuick(j, 1));
			
			if (ded > temperature) {
				ded = temperature / ded;
				dxdy.setQuick(j, 0, dxdy.getQuick(j, 0) * ded);
				dxdy.setQuick(j, 1, dxdy.getQuick(j, 1) * ded);
			}
			
			p1 = transform(v);
			
			p1.setLocation(p1.getX() + dxdy.getQuick(j, 0), p1.getY() + dxdy.getQuick(j, 1));
			
		}
		 break;
			}
			catch (ConcurrentModificationException cme) {}
		}
		       
		 
		

	}

	@Override
	public void initialize() {
		doInit();
		
	}

	@Override
	public void reset() {
		doInit();
		
	}
	
	
	/**
     * Sets the maximum number of iterations.
     */
    public void setMaxIterations(int maxIterations) {
        this.maxIterations = maxIterations;
    }
    
    /**
     * This one is an incremental visualization.
     */
    public boolean isIncremental() {
        return true;
    }
    
    private void doInit() {
    	Graph<NSG.NSGVertex, NSG.NSGEdge> graph = getGraph();
    	Dimension d = getSize();
    	if(graph != null && d != null) {
    		currentIteration = 0;
    		
    		vertexIds = new HashMap<NSG.NSGVertex, Integer>(graph.getVertexCount());
    		
    		int i = 0;
    		
    		for (NSG.NSGVertex v : graph.getVertices()) {
    			vertexIds.put(v, i++);
    		}
    		
    		dxdy = DoubleFactory2D.dense.make(vertexIds.size(), 2);
    		
    		forceConstant = 
    			Math
    			.sqrt(d.getHeight()
    					* d.getWidth()
    					/ graph.getVertexCount());

    		repulseRadius = 800000 * graph.getVertexCount();
    	}
    }
    
    /**
     * Sets the attraction multiplier.
     */
	public void setAttractionMultiplier(double attraction) {
        this.attraction_multiplier = attraction;
    }
    
    /**
     * Sets the repulsion multiplier.
     */
    public void setRepulsionMultiplier(double repulsion) {
        this.repulsion_multiplier = repulsion;
    }


}
